/*******************************************************************************

    Copyright(C) Brian Paul 1999, 2000, 2001.

    Permission is hereby granted, free of charge, to any person obtaining a
    copy of this software and associated documentation files (the "Software"),
    to deal in the Software without restriction, including without limitation
    the rights to use, copy, modify, merge, publish, distribute, sublicense,
    and/or sell copies of the Software, and to permit persons to whom the
    Software is furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    BRIAN PAUL BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
    AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
    CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

    gears.cpp
    Classic OpenGL test program that render colored rotating gears.

*******************************************************************************/

#include <sys/keycodes.h>
#include <sys/termmode.h>

#include <errno.h>
#include <error.h>
#include <fcntl.h>
#include <locale.h>
#include <math.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <timespec.h>
#include <unistd.h>

#include <dispd.h>

#include <GL/gl.h>
#include <GL/osmesa.h>

//#define __sortix_crazy_display_server__

#ifdef __sortix_crazy_display_server__
#include <sys/socket.h>
#include <sys/un.h>

#include <ioleast.h>

#include "../../display/display-protocol.h"

int open_local_client_socket(const char* path, int flags)
{
	size_t path_length = strlen(path);
	size_t addr_size = offsetof(struct sockaddr_un, sun_path) + path_length + 1;
	struct sockaddr_un* sockaddr = (struct sockaddr_un*) malloc(addr_size);
	if ( !sockaddr )
		return -1;
	sockaddr->sun_family = AF_LOCAL;
	strcpy(sockaddr->sun_path, path);
	int fd = socket(AF_LOCAL, SOCK_STREAM | flags, 0);
	if ( fd < 0 )
		return free(sockaddr), -1;
	if ( connect(fd, (const struct sockaddr*) sockaddr, addr_size) < 0 )
		return close(fd), free(sockaddr), -1;
	free(sockaddr);
	return fd;
}

void writeall_or_die(int fd, const void* buffer, size_t count)
{
	size_t amount = writeall(fd, buffer, count);
	if ( amount != count )
		error(1, errno, "writeall");
}

void send_message(int fd, uint32_t message_id,
                  const void* message, size_t message_size,
                  const void* auxilerary, size_t auxilerary_size)
{
	struct display_packet_header header;
	header.message_id = message_id;
	header.message_length = message_size + auxilerary_size;
	writeall_or_die(fd, &header, sizeof(header));
	writeall_or_die(fd, message, message_size);
	writeall_or_die(fd, auxilerary, auxilerary_size);
}

void send_message_no_aux(int fd, uint32_t message_id,
                         const void* message, size_t message_size)
{
	send_message(fd, message_id, message, message_size, 0, 0);
}
#endif

static void gear(GLfloat inner_radius, GLfloat outer_radius, GLfloat width,
                 GLint teeth, GLfloat tooth_depth)
{
	GLint i;
	GLfloat r0, r1, r2;
	GLfloat angle, da;
	GLfloat u, v, len;

	r0 = inner_radius;
	r1 = outer_radius - tooth_depth / 2.0;
	r2 = outer_radius + tooth_depth / 2.0;

	da = 2.0 * M_PI / teeth / 4.0;

	glShadeModel(GL_FLAT);

	glNormal3f(0.0, 0.0, 1.0);

	// Draw the front face.
	glBegin(GL_QUAD_STRIP);
	for ( i = 0; i <= teeth; i++ )
	{
		angle = i * 2.0 * M_PI / teeth;
		glVertex3f(r0 * cos(angle), r0 * sin(angle), width * 0.5);
		glVertex3f(r1 * cos(angle), r1 * sin(angle), width * 0.5);
		if ( i < teeth )
		{
			glVertex3f(r0 * cos(angle), r0 * sin(angle), width * 0.5);
			glVertex3f(r1 * cos(angle + 3 * da), r1 * sin(angle + 3 * da),
			           width * 0.5);
		}
	}
	glEnd();

	// Draw the front sides of teeth.
	glBegin(GL_QUADS);
	da = 2.0 * M_PI / teeth / 4.0;
	for ( i = 0; i < teeth; i++ )
	{
		angle = i * 2.0 * M_PI / teeth;
		glVertex3f(r1 * cos(angle), r1 * sin(angle), width * 0.5);
		glVertex3f(r2 * cos(angle + da), r2 * sin(angle + da), width * 0.5);
		glVertex3f(r2 * cos(angle + 2 * da), r2 * sin(angle + 2 * da),
		           width * 0.5);
		glVertex3f(r1 * cos(angle + 3 * da), r1 * sin(angle + 3 * da),
		           width * 0.5);
	}
	glEnd();

	glNormal3f(0.0, 0.0, -1.0);

	// Draw the back face.
	glBegin(GL_QUAD_STRIP);
	for ( i = 0; i <= teeth; i++ )
	{
		angle = i * 2.0 * M_PI / teeth;
		glVertex3f(r1 * cos(angle), r1 * sin(angle), -width * 0.5);
		glVertex3f(r0 * cos(angle), r0 * sin(angle), -width * 0.5);
		if ( i < teeth )
		{
			glVertex3f(r1 * cos(angle + 3 * da), r1 * sin(angle + 3 * da),
			           -width * 0.5);
			glVertex3f(r0 * cos(angle), r0 * sin(angle), -width * 0.5);
		}
	}
	glEnd();

	// Draw the back sides of teeth.
	glBegin(GL_QUADS);
	da = 2.0 * M_PI / teeth / 4.0;
	for ( i = 0; i < teeth; i++ )
	{
		angle = i * 2.0 * M_PI / teeth;

		glVertex3f(r1 * cos(angle + 3 * da), r1 * sin(angle + 3 * da),
		           -width * 0.5);
		glVertex3f(r2 * cos(angle + 2 * da), r2 * sin(angle + 2 * da),
		           -width * 0.5);
		glVertex3f(r2 * cos(angle + da), r2 * sin(angle + da), -width * 0.5);
		glVertex3f(r1 * cos(angle), r1 * sin(angle), -width * 0.5);
	}
	glEnd();

	// Draw the outward faces of teeth.
	glBegin(GL_QUAD_STRIP);
	for (i = 0; i < teeth; i++)
	{
		angle = i * 2.0 * M_PI / teeth;

		glVertex3f(r1 * cos(angle), r1 * sin(angle), width * 0.5);
		glVertex3f(r1 * cos(angle), r1 * sin(angle), -width * 0.5);
		u = r2 * cos(angle + da) - r1 * cos(angle);
		v = r2 * sin(angle + da) - r1 * sin(angle);
		len = sqrt(u * u + v * v);
		u /= len;
		v /= len;
		glNormal3f(v, -u, 0.0);
		glVertex3f(r2 * cos(angle + da), r2 * sin(angle + da), width * 0.5);
		glVertex3f(r2 * cos(angle + da), r2 * sin(angle + da), -width * 0.5);
		glNormal3f(cos(angle), sin(angle), 0.0);
		glVertex3f(r2 * cos(angle + 2 * da), r2 * sin(angle + 2 * da),
		           width * 0.5);
		glVertex3f(r2 * cos(angle + 2 * da), r2 * sin(angle + 2 * da),
		           -width * 0.5);
		u = r1 * cos(angle + 3 * da) - r2 * cos(angle + 2 * da);
		v = r1 * sin(angle + 3 * da) - r2 * sin(angle + 2 * da);
		glNormal3f(v, -u, 0.0);
		glVertex3f(r1 * cos(angle + 3 * da), r1 * sin(angle + 3 * da),
		           width * 0.5);
		glVertex3f(r1 * cos(angle + 3 * da), r1 * sin(angle + 3 * da),
		           -width * 0.5);
		glNormal3f(cos(angle), sin(angle), 0.0);
	}

	glVertex3f(r1 * cos(0), r1 * sin(0), width * 0.5);
	glVertex3f(r1 * cos(0), r1 * sin(0), -width * 0.5);

	glEnd();

	glShadeModel(GL_SMOOTH);

	// Draw the inside radius cylinder.
	glBegin(GL_QUAD_STRIP);
	for ( i = 0; i <= teeth; i++ )
	{
		angle = i * 2.0 * M_PI / teeth;
		glNormal3f(-cos(angle), -sin(angle), 0.0);
		glVertex3f(r0 * cos(angle), r0 * sin(angle), -width * 0.5);
		glVertex3f(r0 * cos(angle), r0 * sin(angle), width * 0.5);
	}
	glEnd();
}

static GLfloat view_rotx = 20.0, view_roty = 30.0, view_rotz = 0.0;
static GLint gear1, gear2, gear3;
static GLfloat angle = 0.0;

static void draw(int width, int height)
{
	GLfloat h = (GLfloat) height / (GLfloat) width;

	glViewport(0, 0, (GLint) width, (GLint) height);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glFrustum(-1.0, 1.0, -h, h, 5.0, 60.0);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glTranslatef(0.0, 0.0, -40.0);

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glPushMatrix();
	glRotatef(view_rotx, 1.0, 0.0, 0.0);
	glRotatef(view_roty, 0.0, 1.0, 0.0);
	glRotatef(view_rotz, 0.0, 0.0, 1.0);

	glPushMatrix();
	glTranslatef(-3.0, -2.0, 0.0);
	glRotatef(angle, 0.0, 0.0, 1.0);
	glCallList(gear1);
	glPopMatrix();

	glPushMatrix();
	glTranslatef(3.1, -2.0, 0.0);
	glRotatef(-2.0 * angle - 9.0, 0.0, 0.0, 1.0);
	glCallList(gear2);
	glPopMatrix();

	glPushMatrix();
	glTranslatef(-3.1, 4.2, 0.0);
	glRotatef(-2.0 * angle - 25.0, 0.0, 0.0, 1.0);
	glCallList(gear3);
	glPopMatrix();

	glPopMatrix();
}

static void init()
{
	static GLfloat pos[4] = { 5.0, 5.0, 10.0, 0.0 };
	static GLfloat red[4] = { 0.8, 0.1, 0.0, 1.0 };
	static GLfloat green[4] = { 0.0, 0.8, 0.2, 1.0 };
	static GLfloat blue[4] = { 0.2, 0.2, 1.0, 1.0 };

	glLightfv(GL_LIGHT0, GL_POSITION, pos);
	glEnable(GL_CULL_FACE);
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	glEnable(GL_DEPTH_TEST);

	// Make the gears.
	gear1 = glGenLists(1);
	glNewList(gear1, GL_COMPILE);
	glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, red);
	gear(1.0, 4.0, 1.0, 20, 0.7);
	glEndList();

	gear2 = glGenLists(1);
	glNewList(gear2, GL_COMPILE);
	glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, green);
	gear(0.5, 2.0, 2.0, 10, 0.7);
	glEndList();

	gear3 = glGenLists(1);
	glNewList(gear3, GL_COMPILE);
	glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, blue);
	gear(1.3, 2.0, 0.5, 10, 0.7);
	glEndList();

	glEnable(GL_NORMALIZE);
}

int main(int argc, char* argv[])
{
	uint8_t* image_data;
	int width;
	int height;

#ifdef __sortix_crazy_display_server__
	const char* server_path = "/run/display";
	int client_fd = open_local_client_socket(server_path, 0);
	if ( client_fd < 0 )
		error(1, errno, "open_local_client_socket: %s", server_path);

	const uint32_t WINDOW_ID = 0;
	const uint32_t WINDOW_WIDTH = 600;
	const uint32_t WINDOW_HEIGHT = 384;
	const char* WINDOW_TITLE = "Gears";

	{
		struct display_create_window msg;
		msg.window_id = WINDOW_ID;
		send_message_no_aux(client_fd, DISPLAY_CREATE_WINDOW, &msg, sizeof(msg));
	}

	{
		struct display_resize_window msg;
		msg.window_id = WINDOW_ID;
		msg.width = WINDOW_WIDTH;
		msg.height = WINDOW_HEIGHT;
		send_message_no_aux(client_fd, DISPLAY_RESIZE_WINDOW, &msg, sizeof(msg));
	}

	{
		struct display_title_window msg;
		msg.window_id = WINDOW_ID;
		send_message(client_fd, DISPLAY_TITLE_WINDOW, &msg, sizeof(msg), WINDOW_TITLE, strlen(WINDOW_TITLE));
	}

	uint32_t* framebuffer = new uint32_t[WINDOW_WIDTH * WINDOW_HEIGHT];
	memset(framebuffer, 0, sizeof(framebuffer[0]) * WINDOW_WIDTH * WINDOW_HEIGHT);
#else
	if ( !dispd_initialize(&argc, &argv) )
		error(1, 0, "couldn't initialize dispd library");
	struct dispd_session* session = dispd_attach_default_session();
	if ( !session )
		error(1, 0, "couldn't attach to dispd default session");
	if ( !dispd_session_setup_game_rgba(session) )
		error(1, 0, "couldn't setup dispd rgba session");
	struct dispd_window* window = dispd_create_window_game_rgba(session);
	if ( !window )
		error(1, 0, "couldn't create dispd rgba window");
#endif

#ifdef __sortix_crazy_display_server__
#else
	settermmode(0, TERMMODE_KBKEY | TERMMODE_NONBLOCK | TERMMODE_SIGNAL);
#endif

	OSMesaContext gl_ctx = OSMesaCreateContext(OSMESA_BGRA, NULL);

	struct timespec start_ts;
	struct timespec last_ts = timespec_make(0, 0);
	clock_gettime(CLOCK_MONOTONIC, &start_ts);

	bool key_a = false;
	bool key_c = false;
	bool key_d = false;
	bool key_e = false;
	bool key_q = false;
	bool key_s = false;
	bool key_w = false;
	bool key_up = false;
	bool key_down = false;
	bool key_left = false;
	bool key_right = false;
	bool key_pgup = false;
	bool key_pgdown = false;

	bool first_frame = true;
	bool running = true;
	while ( running )
	{
#ifdef __sortix_crazy_display_server__
		width = WINDOW_WIDTH;
		height = WINDOW_HEIGHT;
		image_data = (uint8_t*) framebuffer;
#else
		struct dispd_framebuffer* fb = dispd_begin_render(window);
		if ( !fb )
			error(0, 0, "unable to begin rendering dispd window");
		width = dispd_get_framebuffer_width(fb);
		height = dispd_get_framebuffer_height(fb);
		image_data = dispd_get_framebuffer_data(fb);
#endif

		struct timespec now_ts;
		clock_gettime(CLOCK_MONOTONIC, &now_ts);
		struct timespec current_ts = timespec_sub(now_ts, start_ts);
		float current_time = current_ts.tv_sec + current_ts.tv_nsec / 1000000000.0;
		struct timespec delta_ts = timespec_sub(current_ts, last_ts);
		float delta_time = delta_ts.tv_sec + delta_ts.tv_nsec / 1000000000.0;

#ifdef __sortix_crazy_display_server__
#else
		uint32_t codepoint;
		while ( read(0, &codepoint, sizeof(codepoint)) == sizeof(codepoint) )
		{
			int kbkey = KBKEY_DECODE(codepoint);
			if ( !kbkey )
				continue;
			unsigned int abskbkey = kbkey < 0 ? - (unsigned int) kbkey :
			                                      (unsigned int) kbkey;
			switch ( abskbkey )
			{
			case KBKEY_A: key_a = 0 < kbkey; break;
			case KBKEY_C: key_c = 0 < kbkey; break;
			case KBKEY_D: key_d = 0 < kbkey; break;
			case KBKEY_E: key_e = 0 < kbkey; break;
			case KBKEY_Q: key_q = 0 < kbkey; break;
			case KBKEY_S: key_s = 0 < kbkey; break;
			case KBKEY_W: key_w = 0 < kbkey; break;
			case KBKEY_UP: key_up = 0 < kbkey; break;
			case KBKEY_DOWN: key_down = 0 < kbkey; break;
			case KBKEY_LEFT: key_left = 0 < kbkey; break;
			case KBKEY_RIGHT: key_right = 0 < kbkey; break;
			case KBKEY_PGUP: key_pgup = 0 < kbkey; break;
			case KBKEY_PGDOWN: key_pgdown = 0 < kbkey; break;
			default: break;
			}
		}
#endif

		angle = current_time * 100;

		if ( key_a || key_left )
			view_roty += 90.0 * delta_time;
		if ( key_d || key_right )
			view_roty -= 90.0 * delta_time;
		if ( key_w || key_up )
			view_rotx += 90.0 * delta_time;
		if ( key_s || key_down )
			view_rotx -= 90.0 * delta_time;
		if ( key_q || key_pgup )
			view_rotz += 90.0 * delta_time;
		if ( key_e || key_pgdown )
			view_rotz -= 90.0 * delta_time;
		if ( key_c )
			running = false;

		if ( !OSMesaMakeCurrent(gl_ctx, image_data, GL_UNSIGNED_BYTE, width,
		                        height) )
			error(1, errno, "`OSMesaMakeCurrent'");

		if ( first_frame )
			init();

		draw(width, height);

#ifdef __sortix_crazy_display_server__
		{
			struct display_render_window msg;
			msg.window_id = WINDOW_ID;
			msg.left = 0;
			msg.top = 0;
			msg.width = WINDOW_WIDTH;
			msg.height = WINDOW_HEIGHT;
			send_message(client_fd, DISPLAY_RENDER_WINDOW, &msg, sizeof(msg),
				         framebuffer, sizeof(framebuffer[0]) * WINDOW_WIDTH * WINDOW_HEIGHT);
		}

		if ( first_frame )
		{
			struct display_show_window msg;
			msg.window_id = WINDOW_ID;
			send_message_no_aux(client_fd, DISPLAY_SHOW_WINDOW, &msg, sizeof(msg));
		}
#else
		dispd_finish_render(fb);
#endif

		last_ts = current_ts;

		first_frame = false;
	}

	OSMesaDestroyContext(gl_ctx);

#ifdef __sortix_crazy_display_server__
#else
	dispd_destroy_window(window);
	dispd_detach_session(session);
#endif

	return 0;
}
